#include <iostream>
#include <verilated.h>
#include "VCACHE.h"

using namespace std;

// Current simulation time (64-bit unsigned)
vluint64_t main_time = 0;
vluint64_t max_time = 300;

vluint64_t t_reset = 0;
vluint64_t d_reset = 5;

vluint64_t t_unlock = 10;
vluint64_t d_unlock = 5;

vluint64_t s_req_finished= 0;

// READ ACCESS ADDR_0 (MISS)
vluint64_t t_cache_req_1 = 50;
vluint64_t d_cache_req_1 = 40;

// READ ACCESS ADDR_1 (MISS)
vluint64_t t_cache_req_2 = 100;
vluint64_t d_cache_req_2 = 40;

// READ ACCESS ADDR_0 (HIT)
vluint64_t t_cache_req_3 = 150;
vluint64_t d_cache_req_3 = 40;

// WRITE ACCESS ADDR_1 (HIT)
vluint64_t t_cache_req_4 = 200;
vluint64_t d_cache_req_4 = 40;

// WRITE ACCESS ADDR_1 (MISS)
vluint64_t t_cache_req_5 = 250;
vluint64_t d_cache_req_5 = 40;


vluint64_t t_l2_cache = 0;
vluint64_t d_l2_cache = 20;
vluint64_t h_l2_cache = 1;

// Called by $time in Verilog
double sc_time_stamp() {
    return main_time;  // Note does conversion to real, to match SystemC
}

int main(int argc, char** argv, char** env){
  
  // Prevent unused variable warnings
  if (false && argc && argv && env) {}

  // Set debug level, 0 is off, 9 is highest presently used
  // May be overridden by commandArgs
  Verilated::debug(0);

  // Randomization reset policy
  // May be overridden by commandArgs
  Verilated::randReset(2);

  // Verilator must compute traced signals
  Verilated::traceEverOn(true);

  // Pass arguments so Verilated code can see them, e.g. $value$plusargs
  // This needs to be called before you create any model
  Verilated::commandArgs(argc, argv);

  // Create logs/ directory in case we have traces to put under it
  Verilated::mkdir("logs");

  // Construct the Verilated model, from Vtop.h generated from Verilating "top.v"
  VCACHE* top = new VCACHE();

  top->res_n = !1;
  top->clk = 0;

  // LOCK/UNLOCK
  top->lock = 0;
  top->unlock = 0;
  top->key = 0x00000000;

  // CPU-INTERFACE
  top->phys_addr_in = 0x0000000000;
  top->phys_addr_in_vld = 0;

  // MEM-INTERFACE
  top->data_from_mem = 0x0000000000000000;
  top->data_from_mem_vld = 0;
    
  while(!Verilated::gotFinish()){

    main_time++;

    // toggle clock
    top->clk=!top->clk;

    if(!top->clk){

      if((main_time >= t_reset) && (main_time < (t_reset+d_reset))){
	top->res_n = !1;
      } else {
	top->res_n = !0;
      }

      if((main_time >= t_unlock) && (main_time < (t_unlock+d_unlock))){       
	top->unlock = 1;
	top->key = 0xFFFFFFFF;
      } else {
	top->unlock = 0;
	top->key = 0x00000000;
      }

      if(main_time == t_cache_req_1){ s_req_finished = 0; }
      
      if((main_time >= t_cache_req_1) && (main_time < (t_cache_req_1+d_cache_req_1))){
	
	// CPU-INTERFACE
	if(1==top->data_to_cpu_vld || (1==s_req_finished)){
	  top->phys_addr_in = 0x0000000000;
	  top->phys_addr_in_vld = 0;
	  top->data_from_mem = 0x0000000000;
	  top->data_from_mem_vld = 0;	  
	  s_req_finished = 1;
	} else {
	  top->phys_addr_in = 0xFFFFFFFFFF;
	  top->phys_addr_in_vld = 1;
	}

	
	if(1==top->miss){
	  if((main_time >= t_cache_req_1+d_l2_cache) && (main_time < t_cache_req_1+d_l2_cache+h_l2_cache)){
	    top->data_from_mem = 0x0123456789ABCDEF;
	    top->data_from_mem_vld = 1;
	  } else {
	    top->data_from_mem = 0x0000000000000000;
	    top->data_from_mem_vld = 0;
	  }
	}

      }

      if(main_time == t_cache_req_2){ s_req_finished = 0; }

      if((main_time >= t_cache_req_2) && (main_time < (t_cache_req_2+d_cache_req_2))){
	
	// CPU-INTERFACE
	if(1==top->data_to_cpu_vld || (1==s_req_finished)){
	  top->phys_addr_in = 0x0000000000;
	  top->phys_addr_in_vld = 0;
	  top->data_from_mem = 0x0000000000;
	  top->data_from_mem_vld = 0;	  
	  s_req_finished = 1;
	} else {	// CPU-INTERFACE
	  top->phys_addr_in = 0xAAAAAA803F;
	  top->phys_addr_in_vld = 1;
	}

	if(1==top->miss){
	  if((main_time >= t_cache_req_2+d_l2_cache) && (main_time < t_cache_req_2+d_l2_cache+h_l2_cache)){
	    top->data_from_mem = 0xFEDCBA9876543210;
	    top->data_from_mem_vld = 1;
	  } else {
	    top->data_from_mem = 0x0000000000000000;
	    top->data_from_mem_vld = 0;
	  }
	}	
	
      }

      if(main_time == t_cache_req_3){ s_req_finished = 0; }

      if((main_time >= t_cache_req_3) && (main_time < (t_cache_req_3+d_cache_req_3))){

	// CPU-INTERFACE
	if(1==top->data_to_cpu_vld || (1==s_req_finished)){
	  top->phys_addr_in = 0x0000000000;
	  top->phys_addr_in_vld = 0;
	  top->data_from_mem = 0x0000000000;
	  top->data_from_mem_vld = 0;	  
	  s_req_finished = 1;
	} else {	// CPU-INTERFACE
	  top->phys_addr_in = 0xFFFFFFFFFF;
	  top->phys_addr_in_vld = 1;
	}

	if(1==top->miss){
	  if((main_time >= t_cache_req_3+d_l2_cache) && (main_time < t_cache_req_3+d_l2_cache+h_l2_cache)){
	    top->data_from_mem = 0xFEDCBA9876543210;
	    top->data_from_mem_vld = 1;
	  } else {
	    top->data_from_mem = 0x0000000000000000;
	    top->data_from_mem_vld = 0;
	  }
	}	

     }

      if(main_time == t_cache_req_4){ s_req_finished = 0; }

      if((main_time >= t_cache_req_4) && (main_time < (t_cache_req_4+d_cache_req_4))){

	// CPU-INTERFACE
	if(1==top->data_to_cpu_vld || (1==s_req_finished)){
	  top->phys_addr_in = 0x0000000000;
	  top->phys_addr_in_vld = 0;
	  top->data_from_mem = 0x0000000000;
	  top->data_from_mem_vld = 0;	  
	  top->data_from_cpu = 0x0000000000000000;
	  top->data_from_cpu_vld = 0;
	  s_req_finished = 1;
	} else {	// CPU-INTERFACE
	  top->phys_addr_in = 0xFFFFFFFFFF;
	  top->phys_addr_in_vld = 1;
	  top->data_from_cpu = 0x1010101010101010;
	  top->data_from_cpu_vld = 1;
	}

	if(1==top->miss){
	  if((main_time >= t_cache_req_4+d_l2_cache) && (main_time < t_cache_req_4+d_l2_cache+h_l2_cache)){
	    top->data_from_mem = 0x0101010101010101;
	    top->data_from_mem_vld = 1;
	  } else {
	    top->data_from_mem = 0x0000000000000000;
	    top->data_from_mem_vld = 0;
	  }
	}	

     }
      
      if(main_time == t_cache_req_5){ s_req_finished = 0; }

      if((main_time >= t_cache_req_5) && (main_time < (t_cache_req_5+d_cache_req_5))){

	// CPU-INTERFACE
	if(1==top->data_to_cpu_vld || (1==s_req_finished)){
	  top->phys_addr_in = 0x0000000000;
	  top->phys_addr_in_vld = 0;
	  top->data_from_mem = 0x0000000000;
	  top->data_from_mem_vld = 0;	  
	  top->data_from_cpu = 0x0000000000000000;
	  top->data_from_cpu_vld = 0;
	  s_req_finished = 1;
	} else {	// CPU-INTERFACE
	  top->phys_addr_in = 0x00FFFFFFFF;
	  top->phys_addr_in_vld = 1;
	  top->data_from_cpu = 0x2020202020202020;
	  top->data_from_cpu_vld = 1;
	}

	if(1==top->miss){
	  if((main_time >= t_cache_req_5+d_l2_cache) && (main_time < t_cache_req_5+d_l2_cache+h_l2_cache)){
	    top->data_from_mem = 0x0101010101010101;
	    top->data_from_mem_vld = 1;
	  } else {
	    top->data_from_mem = 0x0000000000000000;
	    top->data_from_mem_vld = 0;
	  }
	}	

     }
      
      // END SIM
      if(main_time >= max_time){
	top->lock=1;
	top->unlock=1;
	top->phys_addr_in = 0xFFFFFFFFFF;
	top->phys_addr_in_vld = 1;	
      }
      
    }
           
    // Evaluate model
    // (If you have multiple models being simulated in the same
    // timestep then instead of eval(), call eval_step() on each, then
    // eval_end_step() on each.)
    top->eval();
    
    // Read outputs
    VL_PRINTF( "[%" VL_PRI64 "d] clk=%x res_n=%x lock=%x unlock=%x key=%4x phys_addr_in=%10lx phys_addr_in_vld=%x -> locked=%x data_to_cpu=%16lx data_to_cpu_vld=%x miss=%x\n",
	       main_time,
	       top->clk,
	       top->res_n,
	       top->lock, top->unlock, top->key,
	       top->phys_addr_in, top->phys_addr_in_vld,
	       top->locked,
	       top->data_to_cpu, top->data_to_cpu_vld,
	       top->miss );             
  }
  
  top->final();

  //  Coverage analysis (since test passed)
#if VM_COVERAGE
  Verilated::mkdir("logs");
  VerilatedCov::write("logs/coverage.dat");
#endif

  // Delete model
  delete top;
  top = nullptr;

  // Exit
  exit(0);
  
}
